# pylint: disable=duplicate-code, broad-exception-caught, too-many-statements, unused-argument, protected-access,  too-many-lines
"""
Module for monitoring VertexAI API calls.
"""

import logging
import math
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.sdk.resources import TELEMETRY_SDK_NAME
from murnitur.lit.__helpers import (
    get_chat_model_cost,
    get_embed_model_cost,
    handle_exception,
)
from murnitur.lit.semcov import SemanticConvetion

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)


def generate_content(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'generate_content' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'generate_content' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'generate_content' method.
            kwargs: Keyword arguments for the 'generate_content' method.

        Returns:
            The response from the original 'generate_content' method.
        """

        # Check if streaming is enabled for the API call
        streaming = kwargs.get("stream", False)

        # pylint: disable=no-else-return
        if streaming:
            # Special handling for streaming response to accommodate the nature of data flow
            def stream_generator():
                with tracer.start_as_current_span(
                    gen_ai_endpoint, kind=SpanKind.CLIENT
                ) as span:
                    # Placeholder for aggregating streaming response
                    llmresponse = ""

                    # Loop through streaming events capturing relevant details
                    for event in wrapped(*args, **kwargs):
                        llmresponse += str(event.text)
                        prompt_tokens = event.usage_metadata.prompt_token_count
                        completion_tokens = event.usage_metadata.candidates_token_count
                        total_tokens = event.usage_metadata.total_token_count
                        yield event

                    # Handling exception ensure observability without disrupting operation
                    try:
                        prompt = str(args[0])

                        model = "/".join(instance._model_name.split("/")[3:])

                        # Calculate cost of the operation
                        cost = get_chat_model_cost(
                            model, pricing_info, prompt_tokens, completion_tokens
                        )

                        # Set Span attributes
                        span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_SYSTEM,
                            SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_TYPE,
                            SemanticConvetion.GEN_AI_TYPE_CHAT,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_REQUEST_MODEL, model
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, True
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                            completion_tokens,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                        )
                        span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                        if trace_content:
                            span.set_attributes(
                                attributes={
                                    SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                                },
                            )
                            span.set_attributes(
                                attributes={
                                    SemanticConvetion.GEN_AI_CONTENT_COMPLETION: llmresponse,
                                },
                            )

                        span.set_status(Status(StatusCode.OK))

                        if disable_metrics is False:
                            attributes = {
                                TELEMETRY_SDK_NAME: "openlit",
                                SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                                SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                                SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                                SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                                SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                            }

                            metrics["genai_requests"].add(1, attributes)
                            metrics["genai_total_tokens"].add(total_tokens, attributes)
                            metrics["genai_completion_tokens"].add(
                                completion_tokens, attributes
                            )
                            metrics["genai_prompt_tokens"].add(
                                prompt_tokens, attributes
                            )
                            metrics["genai_cost"].record(cost, attributes)

                    except Exception as e:
                        handle_exception(span, e)
                        logger.error("Error in trace creation: %s", e)

            return stream_generator()

        # Handling for non-streaming responses
        else:
            with tracer.start_as_current_span(
                gen_ai_endpoint, kind=SpanKind.CLIENT
            ) as span:
                response = wrapped(*args, **kwargs)

                try:
                    # Format 'messages' into a single string
                    prompt = str(args[0])

                    model = "/".join(instance._model_name.split("/")[3:])
                    # Calculate cost of the operation
                    cost = get_chat_model_cost(
                        model,
                        pricing_info,
                        response.usage_metadata.prompt_token_count,
                        response.usage_metadata.candidates_token_count,
                    )

                    # Set Span attribues
                    span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_SYSTEM,
                        SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_TYPE,
                        SemanticConvetion.GEN_AI_TYPE_CHAT,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, False
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS,
                        response.usage_metadata.prompt_token_count,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                        response.usage_metadata.candidates_token_count,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS,
                        response.usage_metadata.total_token_count,
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                    span.set_attribute(SemanticConvetion.GEN_AI_CONTENT_PROMPT, prompt)
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_CONTENT_COMPLETION,
                        response.candidates[0].content.parts[0].text,
                    )
                    if trace_content:
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                            },
                        )
                        span.set_attributes(
                            attributes={
                                # pylint: disable=line-too-long
                                SemanticConvetion.GEN_AI_CONTENT_COMPLETION: response.candidates[
                                    0
                                ]
                                .content.parts[0]
                                .text,
                            },
                        )

                    span.set_status(Status(StatusCode.OK))

                    if disable_metrics is False:
                        attributes = {
                            TELEMETRY_SDK_NAME: "openlit",
                            SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                            SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                            SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                            SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                            SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                        }

                        metrics["genai_requests"].add(1, attributes)
                        metrics["genai_total_tokens"].add(
                            response.usage_metadata.total_token_count, attributes
                        )
                        metrics["genai_completion_tokens"].add(
                            response.usage_metadata.candidates_token_count, attributes
                        )
                        metrics["genai_prompt_tokens"].add(
                            response.usage_metadata.prompt_token_count, attributes
                        )
                        metrics["genai_cost"].record(cost, attributes)

                    # Return original response
                    return response

                except Exception as e:
                    handle_exception(span, e)
                    logger.error("Error in trace creation: %s", e)

                    # Return original response
                    return response

    return wrapper


def send_message(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'generate_content' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'generate_content' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'generate_content' method.
            kwargs: Keyword arguments for the 'generate_content' method.

        Returns:
            The response from the original 'generate_content' method.
        """

        # Check if streaming is enabled for the API call
        streaming = kwargs.get("stream", False)

        # pylint: disable=no-else-return
        if streaming:
            # Special handling for streaming response to accommodate the nature of data flow
            def stream_generator():
                with tracer.start_as_current_span(
                    gen_ai_endpoint, kind=SpanKind.CLIENT
                ) as span:
                    # Placeholder for aggregating streaming response
                    llmresponse = ""

                    # Loop through streaming events capturing relevant details
                    for event in wrapped(*args, **kwargs):
                        llmresponse += str(event.text)
                        prompt_tokens = event.usage_metadata.prompt_token_count
                        completion_tokens = event.usage_metadata.candidates_token_count
                        total_tokens = event.usage_metadata.total_token_count
                        yield event

                    # Handling exception ensure observability without disrupting operation
                    try:
                        prompt = args[0][0]

                        model = "/".join(instance._model._model_name.split("/")[3:])

                        # Calculate cost of the operation
                        cost = get_chat_model_cost(
                            model, pricing_info, prompt_tokens, completion_tokens
                        )

                        # Set Span attributes
                        span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_SYSTEM,
                            SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_TYPE,
                            SemanticConvetion.GEN_AI_TYPE_CHAT,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_REQUEST_MODEL, model
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, True
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                            completion_tokens,
                        )
                        span.set_attribute(
                            SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                        )
                        span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                        if trace_content:
                            span.set_attributes(
                                attributes={
                                    SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                                },
                            )
                            span.set_attributes(
                                attributes={
                                    SemanticConvetion.GEN_AI_CONTENT_COMPLETION: llmresponse,
                                },
                            )

                        span.set_status(Status(StatusCode.OK))

                        if disable_metrics is False:
                            attributes = {
                                TELEMETRY_SDK_NAME: "openlit",
                                SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                                SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                                SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                                SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                                SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                            }

                            metrics["genai_requests"].add(1, attributes)
                            metrics["genai_total_tokens"].add(total_tokens, attributes)
                            metrics["genai_completion_tokens"].add(
                                completion_tokens, attributes
                            )
                            metrics["genai_prompt_tokens"].add(
                                prompt_tokens, attributes
                            )
                            metrics["genai_cost"].record(cost, attributes)

                    except Exception as e:
                        handle_exception(span, e)
                        logger.error("Error in trace creation: %s", e)

            return stream_generator()

        # Handling for non-streaming responses
        else:
            with tracer.start_as_current_span(
                gen_ai_endpoint, kind=SpanKind.CLIENT
            ) as span:
                response = wrapped(*args, **kwargs)

                try:
                    # Format 'messages' into a single string
                    prompt = args[0]

                    model = "/".join(instance._model._model_name.split("/")[3:])
                    # Calculate cost of the operation
                    cost = get_chat_model_cost(
                        model,
                        pricing_info,
                        response.usage_metadata.prompt_token_count,
                        response.usage_metadata.candidates_token_count,
                    )

                    # Set Span attribues
                    span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_SYSTEM,
                        SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_TYPE,
                        SemanticConvetion.GEN_AI_TYPE_CHAT,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, False
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS,
                        response.usage_metadata.prompt_token_count,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                        response.usage_metadata.candidates_token_count,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS,
                        response.usage_metadata.total_token_count,
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                    if trace_content:
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                            },
                        )
                        span.set_attributes(
                            attributes={
                                # pylint: disable=line-too-long
                                SemanticConvetion.GEN_AI_CONTENT_COMPLETION: response.candidates[
                                    0
                                ]
                                .content.parts[0]
                                .text,
                            },
                        )

                    span.set_status(Status(StatusCode.OK))

                    if disable_metrics is False:
                        attributes = {
                            TELEMETRY_SDK_NAME: "openlit",
                            SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                            SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                            SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                            SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                            SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                        }

                        metrics["genai_requests"].add(1, attributes)
                        metrics["genai_total_tokens"].add(
                            response.usage_metadata.total_token_count, attributes
                        )
                        metrics["genai_completion_tokens"].add(
                            response.usage_metadata.candidates_token_count, attributes
                        )
                        metrics["genai_prompt_tokens"].add(
                            response.usage_metadata.prompt_token_count, attributes
                        )
                        metrics["genai_cost"].record(cost, attributes)

                    # Return original response
                    return response

                except Exception as e:
                    handle_exception(span, e)
                    logger.error("Error in trace creation: %s", e)

                    # Return original response
                    return response

    return wrapper


def predict(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'predict' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'predict' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'predict' method.
            kwargs: Keyword arguments for the 'predict' method.

        Returns:
            The response from the original 'predict' method.
        """

        with tracer.start_as_current_span(
            gen_ai_endpoint, kind=SpanKind.CLIENT
        ) as span:
            response = wrapped(*args, **kwargs)

            try:
                prompt = args[0]

                model = instance._model_id
                # pylint: disable=line-too-long
                prompt_tokens = response._prediction_response.metadata["tokenMetadata"][
                    "inputTokenCount"
                ]["totalTokens"]
                completion_tokens = response._prediction_response.metadata[
                    "tokenMetadata"
                ]["outputTokenCount"]["totalTokens"]
                total_tokens = prompt_tokens + completion_tokens

                # Calculate cost of the operation
                cost = get_chat_model_cost(
                    model, pricing_info, prompt_tokens, completion_tokens
                )

                # Set Span attribues
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(
                    SemanticConvetion.GEN_AI_SYSTEM,
                    SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_TYPE, SemanticConvetion.GEN_AI_TYPE_CHAT
                )
                span.set_attribute(SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint)
                span.set_attribute(SemanticConvetion.GEN_AI_ENVIRONMENT, environment)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                )
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, False)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS, completion_tokens
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                )
                span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                if trace_content:
                    span.set_attributes(
                        attributes={
                            SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                        },
                    )
                    span.set_attributes(
                        attributes={
                            SemanticConvetion.GEN_AI_CONTENT_COMPLETION: response.text,
                        },
                    )

                span.set_status(Status(StatusCode.OK))

                if disable_metrics is False:
                    attributes = {
                        TELEMETRY_SDK_NAME: "openlit",
                        SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                        SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                        SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                        SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                        SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                    }

                    metrics["genai_requests"].add(1, attributes)
                    metrics["genai_total_tokens"].add(total_tokens, attributes)
                    metrics["genai_completion_tokens"].add(
                        completion_tokens, attributes
                    )
                    metrics["genai_prompt_tokens"].add(prompt_tokens, attributes)
                    metrics["genai_cost"].record(cost, attributes)

                # Return original response
                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper


def predict_streaming(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'predict' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'predict' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'predict' method.
            kwargs: Keyword arguments for the 'predict' method.

        Returns:
            The response from the original 'predict' method.
        """

        # Special handling for streaming response to accommodate the nature of data flow
        def stream_generator():
            with tracer.start_as_current_span(
                gen_ai_endpoint, kind=SpanKind.CLIENT
            ) as span:
                # Placeholder for aggregating streaming response
                llmresponse = ""

                # Loop through streaming events capturing relevant details
                for event in wrapped(*args, **kwargs):
                    llmresponse += str(event)
                    yield event

                # Handling exception ensure observability without disrupting operation
                try:
                    prompt = args[0]
                    llmresponse = llmresponse.split(
                        "TextGenerationResponse", maxsplit=1
                    )[0].rstrip()

                    prompt_tokens = math.ceil(len(prompt) / 4)
                    completion_tokens = math.ceil(len(llmresponse) / 4)
                    total_tokens = prompt_tokens + completion_tokens

                    model = instance._model_id

                    # Calculate cost of the operation
                    cost = get_chat_model_cost(
                        model, pricing_info, prompt_tokens, completion_tokens
                    )

                    # Set Span attributes
                    span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_SYSTEM,
                        SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_TYPE,
                        SemanticConvetion.GEN_AI_TYPE_CHAT,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, True)
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                        completion_tokens,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                    if trace_content:
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                            },
                        )
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_COMPLETION: llmresponse,
                            },
                        )

                    span.set_status(Status(StatusCode.OK))

                    if disable_metrics is False:
                        attributes = {
                            TELEMETRY_SDK_NAME: "openlit",
                            SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                            SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                            SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                            SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                            SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                        }

                        metrics["genai_requests"].add(1, attributes)
                        metrics["genai_total_tokens"].add(total_tokens, attributes)
                        metrics["genai_completion_tokens"].add(
                            completion_tokens, attributes
                        )
                        metrics["genai_prompt_tokens"].add(prompt_tokens, attributes)
                        metrics["genai_cost"].record(cost, attributes)

                except Exception as e:
                    handle_exception(span, e)
                    logger.error("Error in trace creation: %s", e)

        return stream_generator()

    return wrapper


def start_chat(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'start_chat' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'start_chat' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'start_chat' method.
            kwargs: Keyword arguments for the 'start_chat' method.

        Returns:
            The response from the original 'start_chat' method.
        """

        with tracer.start_as_current_span(
            gen_ai_endpoint, kind=SpanKind.CLIENT
        ) as span:
            response = wrapped(*args, **kwargs)

            try:
                prompt = args[0]

                model = instance._model._model_id

                # pylint: disable=line-too-long
                prompt_tokens = response._prediction_response.metadata["tokenMetadata"][
                    "inputTokenCount"
                ]["totalTokens"]
                completion_tokens = response._prediction_response.metadata[
                    "tokenMetadata"
                ]["outputTokenCount"]["totalTokens"]
                total_tokens = prompt_tokens + completion_tokens

                # Calculate cost of the operation
                cost = get_chat_model_cost(
                    model, pricing_info, prompt_tokens, completion_tokens
                )

                # Set Span attribues
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(
                    SemanticConvetion.GEN_AI_SYSTEM,
                    SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_TYPE, SemanticConvetion.GEN_AI_TYPE_CHAT
                )
                span.set_attribute(SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint)
                span.set_attribute(SemanticConvetion.GEN_AI_ENVIRONMENT, environment)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                )
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, False)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS, completion_tokens
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                )
                span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                if trace_content:
                    span.set_attributes(
                        attributes={
                            SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                        },
                    )
                    span.set_attributes(
                        attributes={
                            SemanticConvetion.GEN_AI_CONTENT_COMPLETION: response.text,
                        },
                    )

                span.set_status(Status(StatusCode.OK))

                if disable_metrics is False:
                    attributes = {
                        TELEMETRY_SDK_NAME: "openlit",
                        SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                        SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                        SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                        SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                        SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                    }

                    metrics["genai_requests"].add(1, attributes)
                    metrics["genai_total_tokens"].add(total_tokens, attributes)
                    metrics["genai_completion_tokens"].add(
                        completion_tokens, attributes
                    )
                    metrics["genai_prompt_tokens"].add(prompt_tokens, attributes)
                    metrics["genai_cost"].record(cost, attributes)

                # Return original response
                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper


def start_chat_streaming(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'start_chat' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'start_chat' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'start_chat' method.
            kwargs: Keyword arguments for the 'start_chat' method.

        Returns:
            The response from the original 'start_chat' method.
        """

        # Special handling for streaming response to accommodate the nature of data flow
        def stream_generator():
            with tracer.start_as_current_span(
                gen_ai_endpoint, kind=SpanKind.CLIENT
            ) as span:
                # Placeholder for aggregating streaming response
                llmresponse = ""

                # Loop through streaming events capturing relevant details
                for event in wrapped(*args, **kwargs):
                    llmresponse += str(event.text)
                    yield event

                # Handling exception ensure observability without disrupting operation
                try:
                    prompt = args[0]

                    prompt_tokens = math.ceil(len(prompt) / 4)
                    completion_tokens = math.ceil(len(llmresponse) / 4)
                    total_tokens = prompt_tokens + completion_tokens

                    model = instance._model._model_id

                    # Calculate cost of the operation
                    cost = get_chat_model_cost(
                        model, pricing_info, prompt_tokens, completion_tokens
                    )

                    # Set Span attributes
                    span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_SYSTEM,
                        SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_TYPE,
                        SemanticConvetion.GEN_AI_TYPE_CHAT,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_ENVIRONMENT, environment
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                    span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, True)
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_COMPLETION_TOKENS,
                        completion_tokens,
                    )
                    span.set_attribute(
                        SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, total_tokens
                    )
                    span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                    if trace_content:
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                            },
                        )
                        span.set_attributes(
                            attributes={
                                SemanticConvetion.GEN_AI_CONTENT_COMPLETION: llmresponse,
                            },
                        )

                    span.set_status(Status(StatusCode.OK))

                    if disable_metrics is False:
                        attributes = {
                            TELEMETRY_SDK_NAME: "openlit",
                            SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                            SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                            SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                            SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_CHAT,
                            SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                        }

                        metrics["genai_requests"].add(1, attributes)
                        metrics["genai_total_tokens"].add(total_tokens, attributes)
                        metrics["genai_completion_tokens"].add(
                            completion_tokens, attributes
                        )
                        metrics["genai_prompt_tokens"].add(prompt_tokens, attributes)
                        metrics["genai_cost"].record(cost, attributes)

                except Exception as e:
                    handle_exception(span, e)
                    logger.error("Error in trace creation: %s", e)

        return stream_generator()

    return wrapper


def embeddings(
    gen_ai_endpoint,
    version,
    environment,
    application_name,
    tracer,
    pricing_info,
    trace_content,
    metrics,
    disable_metrics,
):
    """
    Generates a telemetry wrapper for messages to collect metrics.

    Args:
        gen_ai_endpoint: Endpoint identifier for logging and tracing.
        version: Version of the monitoring package.
        environment: Deployment environment (e.g., production, staging).
        application_name: Name of the application using the OpenAI API.
        tracer: OpenTelemetry tracer for creating spans.
        pricing_info: Information used for calculating the cost of OpenAI usage.
        trace_content: Flag indicating whether to trace the actual content.

    Returns:
        A function that wraps the chat method to add telemetry.
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the 'generate_content' API call to add telemetry.

        This collects metrics such as execution time, cost, and token usage, and handles errors
        gracefully, adding details to the trace for observability.

        Args:
            wrapped: The original 'generate_content' method to be wrapped.
            instance: The instance of the class where the original method is defined.
            args: Positional arguments for the 'generate_content' method.
            kwargs: Keyword arguments for the 'generate_content' method.

        Returns:
            The response from the original 'generate_content' method.
        """

        with tracer.start_as_current_span(
            gen_ai_endpoint, kind=SpanKind.CLIENT
        ) as span:
            response = wrapped(*args, **kwargs)

            try:
                prompt = args[0][0]

                model = instance._model_id

                prompt_tokens = int(response[0].statistics.token_count)

                # Calculate cost of the operation
                cost = get_embed_model_cost(model, pricing_info, prompt_tokens)

                # Set Span attribues
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(
                    SemanticConvetion.GEN_AI_SYSTEM,
                    SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_TYPE,
                    SemanticConvetion.GEN_AI_TYPE_EMBEDDING,
                )
                span.set_attribute(SemanticConvetion.GEN_AI_ENDPOINT, gen_ai_endpoint)
                span.set_attribute(SemanticConvetion.GEN_AI_ENVIRONMENT, environment)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_APPLICATION_NAME, application_name
                )
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_MODEL, model)
                span.set_attribute(SemanticConvetion.GEN_AI_REQUEST_IS_STREAM, False)
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_PROMPT_TOKENS, prompt_tokens
                )
                span.set_attribute(
                    SemanticConvetion.GEN_AI_USAGE_TOTAL_TOKENS, prompt_tokens
                )
                span.set_attribute(SemanticConvetion.GEN_AI_USAGE_COST, cost)
                if trace_content:
                    span.set_attributes(
                        attributes={
                            SemanticConvetion.GEN_AI_CONTENT_PROMPT: prompt,
                        },
                    )

                span.set_status(Status(StatusCode.OK))

                if disable_metrics is False:
                    attributes = {
                        TELEMETRY_SDK_NAME: "openlit",
                        SemanticConvetion.GEN_AI_APPLICATION_NAME: application_name,
                        SemanticConvetion.GEN_AI_SYSTEM: SemanticConvetion.GEN_AI_SYSTEM_VERTEXAI,
                        SemanticConvetion.GEN_AI_ENVIRONMENT: environment,
                        SemanticConvetion.GEN_AI_TYPE: SemanticConvetion.GEN_AI_TYPE_EMBEDDING,
                        SemanticConvetion.GEN_AI_REQUEST_MODEL: model,
                    }

                    metrics["genai_requests"].add(1, attributes)
                    metrics["genai_total_tokens"].add(prompt_tokens, attributes)
                    metrics["genai_prompt_tokens"].add(prompt_tokens, attributes)
                    metrics["genai_cost"].record(cost, attributes)

                # Return original response
                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper



# pylint: disable=useless-return, bad-staticmethod-argument, disable=duplicate-code
"""Initializer of Auto Instrumentation of AWS Bedrock Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from murnitur.lit.instrumentation.bedrock.bedrock import converse

_instruments = ("boto3 >= 1.34.138",)


class BedrockInstrumentor(BaseInstrumentor):
    """
    An instrumentor for AWS Bedrock's client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default_application")
        environment = kwargs.get("environment", "default_environment")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        trace_content = kwargs.get("trace_content", False)
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("boto3")

        # sync
        wrap_function_wrapper(
            "botocore.client",
            "ClientCreator.create_client",
            converse(
                "bedrock.converse",
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                trace_content,
                metrics,
                disable_metrics,
            ),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from murnitur.instrumentation.astradb.astradb import general_wrap

_instruments = ("astrapy >= 1.3.1",)

WRAPPED_METHODS = [
    {
        "package": "astrapy.db",
        "object": "AstraDB.create_collection",
        "endpoint": "astradb.create_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDB.delete_collection",
        "endpoint": "astradb.delete_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.insert_one",
        "endpoint": "astradb.insert_one",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.insert_many",
        "endpoint": "astradb.insert_many",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.delete_one",
        "endpoint": "astradb.delete_one",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.delete_many",
        "endpoint": "astradb.delete_many",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.update_one",
        "endpoint": "astradb.update_one",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.update_many",
        "endpoint": "astradb.update_many",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.upsert_one",
        "endpoint": "astradb.upsert_one",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.upsert_many",
        "endpoint": "astradb.upsert_many",
        "wrapper": general_wrap,
    },
    {
        "package": "astrapy.db",
        "object": "AstraDBCollection.find",
        "endpoint": "astradb.find",
        "wrapper": general_wrap,
    },
]


class AstraDBInstrumentor(BaseInstrumentor):
    """An instrumentor for AstraDB's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name")
        environment = kwargs.get("environment")
        tracer = kwargs.get("tracer")
        version = importlib.metadata.version("astrapy")

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            gen_ai_endpoint = wrapped_method.get("endpoint")
            wrapper = wrapped_method.get("wrapper")
            wrap_function_wrapper(
                wrap_package,
                wrap_object,
                wrapper(
                    gen_ai_endpoint, version, environment, application_name, tracer
                ),
            )

    @staticmethod
    def _uninstrument(self, **kwargs):
        pass



# pylint: disable=useless-return, bad-staticmethod-argument, disable=duplicate-code
"""Initializer of Auto Instrumentation of LlamaIndex Functions"""
from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from murnitur.lit.instrumentation.llamaindex.llamaindex import load_data

_instruments = ("llama-index >= 0.10.0",)

WRAPPED_METHODS = [
    {
        "package": "llama_index.core.readers",
        "object": "SimpleDirectoryReader.load_data",
        "endpoint": "llamaindex.load_data",
        "wrapper": load_data,
    },
    {
        "package": "llama_index.core.node_parser",
        "object": "SentenceSplitter.get_nodes_from_documents",
        "endpoint": "llamaindex.data_splitter",
        "wrapper": load_data,
    },
]


class LlamaIndexInstrumentor(BaseInstrumentor):
    """An instrumentor for LlamaIndex's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name")
        environment = kwargs.get("environment")
        tracer = kwargs.get("tracer")
        pricing_info = kwargs.get("pricing_info")
        trace_content = kwargs.get("trace_content")
        version = importlib.metadata.version("llama-index")

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            gen_ai_endpoint = wrapped_method.get("endpoint")
            wrapper = wrapped_method.get("wrapper")
            wrap_function_wrapper(
                wrap_package,
                wrap_object,
                wrapper(
                    gen_ai_endpoint,
                    version,
                    environment,
                    application_name,
                    tracer,
                    pricing_info,
                    trace_content,
                ),
            )

    @staticmethod
    def _uninstrument(self, **kwargs):
        pass



import os
import json
from pydantic import BaseModel, Field
from typing import Any, Optional, List, Dict, Union
import shutil
import webbrowser
import sys
import datetime
import portalocker
from rich.table import Table
from rich.console import Console
from rich import print

from murnitur.annotation.evaluation.metrics import BaseMetric
from murnitur.annotation.evaluation.api import Api, Endpoints
from murnitur.annotation.evaluation.test_run.api import (
    LLMApiTestCase,
    ConversationalApiTestCase,
    TestRunHttpResponse,
)
from murnitur.annotation.evaluation.test_case import LLMTestCase, ConversationalTestCase
from murnitur.annotation.evaluation.utils import (
    delete_file_if_exists,
    get_is_running_deepeval,
    is_confident,
)
from murnitur.annotation.evaluation.test_run.cache import test_run_cache_manager

TEMP_FILE_NAME = "temp_test_run_data.json"


class MetricScoreType(BaseModel):
    metric: str
    score: float

    @classmethod
    def from_metric(cls, metric: BaseMetric):
        return cls(metric=metric.__name__, score=metric.score)


class MetricScores(BaseModel):
    metric: str
    scores: List[float]


class DeploymentConfigs(BaseModel):
    env: str
    actor: Optional[str]
    branch: Optional[str]
    sha: Optional[str]
    repo: Optional[str]


class MetricsAverageDict:
    def __init__(self):
        self.metric_dict = {}
        self.metric_count = {}

    def add_metric(self, metric_name, score):
        if metric_name not in self.metric_dict:
            self.metric_dict[metric_name] = score
            self.metric_count[metric_name] = 1
        else:
            self.metric_dict[metric_name] += score
            self.metric_count[metric_name] += 1

    def get_average_metric_score(self):
        return [
            MetricScoreType(
                metric=metric,
                score=self.metric_dict[metric] / self.metric_count[metric],
            )
            for metric in self.metric_dict
        ]


class RemainingTestRun(BaseModel):
    testRunId: str
    test_cases: List[LLMApiTestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )
    conversational_test_cases: List[ConversationalApiTestCase] = Field(
        alias="conversationalTestCases", default_factory=lambda: []
    )


class TestRun(BaseModel):
    test_file: Optional[str] = Field(
        None,
        alias="testFile",
    )
    deployment: Optional[bool] = Field(True)
    deployment_configs: Optional[DeploymentConfigs] = Field(
        None, alias="deploymentConfigs"
    )
    test_cases: List[LLMApiTestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )
    conversational_test_cases: List[ConversationalApiTestCase] = Field(
        alias="conversationalTestCases", default_factory=lambda: []
    )
    metrics_scores: List[MetricScores] = Field(
        default_factory=lambda: [], alias="metricsScores"
    )
    hyperparameters: Optional[Dict[Any, Any]] = Field(None)
    test_passed: Optional[int] = Field(None, alias="testPassed")
    test_failed: Optional[int] = Field(None, alias="testFailed")
    run_duration: float = Field(0.0, alias="runDuration")
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    dataset_alias: Optional[str] = Field(None, alias="datasetAlias")
    dataset_id: Optional[str] = Field(None, alias="datasetId")

    def add_test_case(
        self, api_test_case: Union[LLMApiTestCase, ConversationalApiTestCase]
    ):
        if isinstance(api_test_case, ConversationalApiTestCase):
            self.conversational_test_cases.append(api_test_case)
        else:
            self.test_cases.append(api_test_case)

        if api_test_case.evaluation_cost is not None:
            if self.evaluation_cost is None:
                self.evaluation_cost = api_test_case.evaluation_cost
            else:
                self.evaluation_cost += api_test_case.evaluation_cost

    def set_dataset_properties(
        self, test_case: Union[LLMTestCase, ConversationalTestCase]
    ):
        if self.dataset_alias is None:
            self.dataset_alias = test_case._dataset_alias

        if self.dataset_id is None:
            self.dataset_id = test_case._dataset_id

    def sort_test_cases(self):
        self.test_cases.sort(
            key=lambda x: (
                x.order if x.order is not None else float("inf"),
                x.name,
            )
        )
        # Optionally update order only if not already set
        highest_order = 0
        for test_case in self.test_cases:
            if test_case.order is None:
                test_case.order = highest_order
            highest_order = test_case.order + 1

        self.conversational_test_cases.sort(
            key=lambda x: (
                x.order if x.order is not None else float("inf"),
                x.name,
            )
        )
        # Optionally update order only if not already set
        highest_order = 0
        for test_case in self.conversational_test_cases:
            if test_case.order is None:
                test_case.order = highest_order
            highest_order = test_case.order + 1

    def construct_metrics_scores(self) -> int:
        metrics_dict: Dict[str, List[float]] = {}
        valid_scores = 0
        for test_case in self.test_cases:
            if test_case.metrics_metadata is None:
                continue
            for metric_metadata in test_case.metrics_metadata:
                metric = metric_metadata.metric
                score = metric_metadata.score
                if score is None:
                    continue
                valid_scores += 1
                if metric in metrics_dict:
                    metrics_dict[metric].append(score)
                else:
                    metrics_dict[metric] = [score]

        for test_case in self.conversational_test_cases:
            # right now, we only look at individual message evaluations
            # and not a conversation as a whole
            for message in test_case.messages:
                if message.metrics_metadata is None:
                    continue
                for metric_metadata in message.metrics_metadata:
                    metric = metric_metadata.metric
                    score = metric_metadata.score
                    if score is None:
                        continue
                    valid_scores += 1
                    if metric in metrics_dict:
                        metrics_dict[metric].append(score)
                    else:
                        metrics_dict[metric] = [score]

        # metrics_scores combines both conversational and nonconvo scores
        # might need to separate in the future
        self.metrics_scores = [
            MetricScores(metric=metric, scores=scores)
            for metric, scores in metrics_dict.items()
        ]
        return valid_scores

    def calculate_test_passes_and_fails(self):
        test_passed = 0
        test_failed = 0
        for test_case in self.test_cases:
            if test_case.success is not None:
                if test_case.success:
                    test_passed += 1
                else:
                    test_failed += 1

        for test_case in self.conversational_test_cases:
            # we don't count for conversational test cases success,
            # because that would be double counting
            for message in test_case.messages:
                if message.success is not None:
                    # check None for messages that are not evaluated
                    if message.success:
                        test_passed += 1
                    else:
                        test_failed += 1

        self.test_passed = test_passed
        self.test_failed = test_failed

    def save(self, f):
        try:
            body = self.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = self.dict(by_alias=True, exclude_none=True)
        json.dump(body, f)
        return self

    @classmethod
    def load(cls, f):
        return cls(**json.load(f))


class TestRunManager:
    def __init__(self):
        self.test_run = None
        self.temp_file_name = TEMP_FILE_NAME
        self.save_to_disk = False
        self.disable_request = False

    def reset(self):
        self.test_run = None
        self.temp_file_name = TEMP_FILE_NAME
        self.save_to_disk = False
        self.disable_request = False

    def set_test_run(self, test_run: TestRun):
        self.test_run = test_run

    def create_test_run(
        self,
        deployment: Optional[bool] = False,
        deployment_configs: Optional[DeploymentConfigs] = None,
        file_name: Optional[str] = None,
        disable_request: Optional[bool] = False,
    ):
        self.disable_request = disable_request
        test_run = TestRun(
            testFile=file_name,
            testCases=[],
            metricsScores=[],
            hyperparameters=None,
            deployment=deployment,
            deploymentConfigs=deployment_configs,
            testPassed=None,
            testFailed=None,
        )
        self.set_test_run(test_run)

        if self.save_to_disk:
            self.save_test_run()

    def get_test_run(self):
        if self.test_run is None:
            self.create_test_run()

        if self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_name,
                    mode="r",
                    flags=portalocker.LOCK_SH | portalocker.LOCK_NB,
                ) as file:
                    self.test_run = self.test_run.load(file)
            except (
                FileNotFoundError,
                portalocker.exceptions.LockException,
            ) as e:
                print(f"Error loading test run from disk: {e}", file=sys.stderr)
                self.test_run = None

        return self.test_run

    def save_test_run(self):
        if self.save_to_disk:
            try:
                with portalocker.Lock(self.temp_file_name, mode="w") as file:
                    self.test_run = self.test_run.save(file)
            except portalocker.exceptions.LockException:
                print(
                    "In save_test_run, Error saving test run to disk",
                    file=sys.stderr,
                )

    def update_test_run(
        self,
        api_test_case: Union[LLMApiTestCase, ConversationalApiTestCase],
        test_case: Union[LLMTestCase, ConversationalTestCase],
    ):
        if self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_name,
                    mode="r+",
                    flags=portalocker.LOCK_EX,
                ) as file:
                    file.seek(0)
                    self.test_run = self.test_run.load(file)

                    # Update the test run object
                    self.test_run.add_test_case(api_test_case)
                    self.test_run.set_dataset_properties(test_case)

                    # Save the updated test run back to the file
                    file.seek(0)
                    file.truncate()
                    self.test_run.save(file)
            except (
                FileNotFoundError,
                portalocker.exceptions.LockException,
            ) as e:
                print(f"Error updating test run to disk: {e}", file=sys.stderr)
                self.test_run = None
        else:
            if self.test_run is None:
                self.create_test_run()

            self.test_run.add_test_case(api_test_case)
            self.test_run.set_dataset_properties(test_case)

    def clear_test_run(self):
        self.test_run = None

    def display_results_table(self, test_run: TestRun):
        table = Table(title="Test Results")
        table.add_column("Test case", justify="left")
        table.add_column("Metric", justify="left")
        table.add_column("Score", justify="left")
        table.add_column("Status", justify="left")
        table.add_column("Overall Success Rate", justify="left")

        for index, test_case in enumerate(test_run.test_cases):
            pass_count = 0
            fail_count = 0
            test_case_name = test_case.name

            for metric_metadata in test_case.metrics_metadata:
                if metric_metadata.success:
                    pass_count += 1
                else:
                    fail_count += 1

            table.add_row(
                test_case_name,
                "",
                "",
                "",
                f"{round((100*pass_count)/(pass_count+fail_count),2)}%",
            )

            for metric_metadata in test_case.metrics_metadata:
                if metric_metadata.error:
                    status = "[red]ERRORED[/red]"
                elif metric_metadata.success:
                    status = "[green]PASSED[/green]"
                else:
                    status = "[red]FAILED[/red]"

                evaluation_model = metric_metadata.evaluation_model
                if evaluation_model is None:
                    evaluation_model = "n/a"

                if metric_metadata.score is not None:
                    metric_score = round(metric_metadata.score, 2)
                else:
                    metric_score = None

                table.add_row(
                    "",
                    str(metric_metadata.metric),
                    f"{metric_score} (threshold={metric_metadata.threshold}, evaluation model={evaluation_model}, reason={metric_metadata.reason}, error={metric_metadata.error})",
                    status,
                    "",
                )

            if index is not len(self.test_run.test_cases) - 1:
                table.add_row(
                    "",
                    "",
                    "",
                    "",
                    "",
                )

        for index, conversattional_test_case in enumerate(
            test_run.conversational_test_cases
        ):
            for test_case in conversattional_test_case.messages:
                if test_case.metrics_metadata is None:
                    # skip if no evaluation
                    continue

                pass_count = 0
                fail_count = 0
                test_case_name = test_case.name

                for metric_metadata in test_case.metrics_metadata:
                    if metric_metadata.success:
                        pass_count += 1
                    else:
                        fail_count += 1

                table.add_row(
                    test_case_name,
                    "",
                    "",
                    "",
                    f"{round((100*pass_count)/(pass_count+fail_count),2)}%",
                )

                for metric_metadata in test_case.metrics_metadata:
                    if metric_metadata.error:
                        status = "[red]ERRORED[/red]"
                    elif metric_metadata.success:
                        status = "[green]PASSED[/green]"
                    else:
                        status = "[red]FAILED[/red]"

                    evaluation_model = metric_metadata.evaluation_model
                    if evaluation_model is None:
                        evaluation_model = "n/a"

                    if metric_metadata.score is not None:
                        metric_score = round(metric_metadata.score, 2)
                    else:
                        metric_score = None

                    table.add_row(
                        "",
                        str(metric_metadata.metric),
                        f"{metric_score} (threshold={metric_metadata.threshold}, evaluation model={evaluation_model}, reason={metric_metadata.reason}, error={metric_metadata.error})",
                        status,
                        "",
                    )

            if index is not len(self.test_run.conversational_test_cases) - 1:
                table.add_row(
                    "",
                    "",
                    "",
                    "",
                    "",
                )

        print(table)
        print(f"Total estimated evaluation tokens cost: {test_run.evaluation_cost} USD")

    def post_test_run(self, test_run: TestRun):
        console = Console()

        if is_confident() and self.disable_request is False:
            BATCH_SIZE = 60
            CONVERSATIONAL_BATCH_SIZE = BATCH_SIZE // 3

            initial_batch = test_run.test_cases[:BATCH_SIZE]
            remaining_test_cases = test_run.test_cases[BATCH_SIZE:]

            initial_conversational_batch = test_run.conversational_test_cases[
                :CONVERSATIONAL_BATCH_SIZE
            ]
            remaining_conversational_test_cases = test_run.conversational_test_cases[
                CONVERSATIONAL_BATCH_SIZE:
            ]

            if (
                len(remaining_test_cases) > 0
                or len(remaining_conversational_test_cases) > 0
            ):
                console.print(
                    "Sending a large test run to Confident, this might take a bit longer than usual..."
                )

            ####################
            ### POST REQUEST ###
            ####################
            test_run.test_cases = initial_batch
            test_run.conversational_test_cases = initial_conversational_batch
            try:
                body = test_run.model_dump(by_alias=True, exclude_none=True)
            except AttributeError:
                # Pydantic version below 2.0
                body = test_run.dict(by_alias=True, exclude_none=True)

            api = Api()
            result = api.post_request(
                endpoint=Endpoints.TEST_RUN_ENDPOINT.value,
                body=body,
            )
            response = TestRunHttpResponse(
                testRunId=result["testRunId"],
                projectId=result["projectId"],
                link=result["link"],
            )
            link = response.link
            ################################################
            ### Send the remaining test cases in batches ###
            ################################################
            max_iterations = (
                max(
                    len(remaining_test_cases),
                    len(remaining_conversational_test_cases),
                )
                // CONVERSATIONAL_BATCH_SIZE
            )
            for i in range(0, max_iterations + 1):
                test_case_index = (
                    i * CONVERSATIONAL_BATCH_SIZE * 3
                )  # Multiply by 3 to match the conversational batch step
                test_case_batch = remaining_test_cases[
                    test_case_index : test_case_index + BATCH_SIZE
                ]

                # Adjusting for conversational_test_cases
                conversational_index = i * CONVERSATIONAL_BATCH_SIZE
                conversational_batch = remaining_conversational_test_cases[
                    conversational_index : conversational_index
                    + CONVERSATIONAL_BATCH_SIZE
                ]

                if len(test_case_batch) == 0 and len(conversational_batch) == 0:
                    break

                remaining_test_run = RemainingTestRun(
                    testRunId=response.testRunId,
                    testCases=test_case_batch,
                    conversationalTestCases=conversational_batch,
                )

                body = None
                try:
                    body = remaining_test_run.model_dump(
                        by_alias=True, exclude_none=True
                    )
                except AttributeError:
                    # Pydantic version below 2.0
                    body = remaining_test_run.dict(by_alias=True, exclude_none=True)

                try:
                    result = api.put_request(
                        endpoint=Endpoints.TEST_RUN_ENDPOINT.value,
                        body=body,
                    )
                except Exception as e:
                    message = f"Unexpected error when sending some test cases. Incomplete test run available at {link}"
                    raise Exception(message) from e

            console.print(
                "✅ Tests finished! View results on " f"[link={link}]{link}[/link]"
            )
            if test_run.deployment == False:
                webbrowser.open(link)

        else:
            console.print(
                '✅ Tests finished! Visit "https://murnitur.ai" to view evaluation results on the web.'
            )

    def save_test_run_locally(self):
        local_folder = os.getenv("DEEPEVAL_RESULTS_FOLDER")
        if local_folder:
            new_test_filename = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            os.rename(self.temp_file_name, new_test_filename)
            if not os.path.exists(local_folder):
                os.mkdir(local_folder)
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            elif os.path.isfile(local_folder):
                print(
                    f"""❌ Error: DEEPEVAL_RESULTS_FOLDER={local_folder} already exists and is a file.\nDetailed results won't be saved. Please specify a folder or an available path."""
                )
            else:
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            os.remove(new_test_filename)

    def wrap_up_test_run(self, runDuration: float, display_table: bool = True):
        test_run = self.get_test_run()
        if test_run is None:
            print("Test Run is empty, please try again.")
            delete_file_if_exists(self.temp_file_name)
            return
        elif (
            len(test_run.test_cases) == 0
            and len(test_run.conversational_test_cases) == 0
        ):
            print("No test cases found, please try again.")
            delete_file_if_exists(self.temp_file_name)
            return

        valid_scores = test_run.construct_metrics_scores()
        if valid_scores == 0:
            print("All metrics errored for all test cases, please try again.")
            delete_file_if_exists(self.temp_file_name)
            delete_file_if_exists(test_run_cache_manager.temp_cache_file_name)
            return
        test_run.run_duration = runDuration
        test_run.calculate_test_passes_and_fails()
        test_run.sort_test_cases()

        if test_run_cache_manager.disable_write_cache is None:
            test_run_cache_manager.disable_write_cache = (
                get_is_running_deepeval() == False
            )

        test_run_cache_manager.wrap_up_cached_test_run()

        if display_table:
            self.display_results_table(test_run)

        self.save_test_run_locally()
        delete_file_if_exists(self.temp_file_name)
        self.post_test_run(test_run)


test_run_manager = TestRunManager()
